/* vmtmconf.c	Initialisation and configuration of VMTM transputer boards
 *
 * Copyright (C) Parsytec Gmbh, 1990
 *
 * SCCSid:	@(#)vmtmconf.c	1.2	94/12/07
 *
 * Author:	Harald Dunkel
 *
 * Date:	30.7.1990	Created
 *
 * Version:	1.0
 *
 * Updates:
 *	Vers	date	author	description
 *
 * Description:
 *	This program initialises the VMTM board.
 *	First a program written in Occam is downloaded to the
 *	NCU.
 *	Second some arrays are built describing the configuration
 *	defined in a configuration file (The configuration must
 *	be defined at compile time).
 *	Third these arrays are sent to the NCU.
 *
 *	See also file vmtm.h
 */

#include <stdio.h>

#include "vmtm.h"

#include "../link_c/link.h"

#define VMTM_BOOT_FILE	"/usr/etc/transp/vmtmconf.cde"


/* C004 control bytes definitions */

#define VMTM_C004_CONNECT	1
#define VMTM_C004_DISCONNECT	6
#define VMTM_C004_END		3
#define VMTM_C004_RESET		4
#define VMTM_C004_READ		2

#define UNUSED 0x7F		/* unused entry in compressed */
				/* connection matrix */


#define	BUFF_SIZE	(4 * 1024)
#define CONFIG_TIMEOUT	40		/* in 1/10 seconds */
#define	WRITE_TIMEOUT	30
#define	READ_TIMEOUT	30
#define	BOOT_TIMEOUT	10		/* in 1/10 seconds */

/* exit codes */
#define	EINFOLINK	1	/* failed to get info about links */
#define	ELOPEN		2	/* failed to open link */
#define	ENOCONFL	3	/* link is no config link */
#define EPARAM          4       /* illegal parameter in argv */

#define OK		1


/*
	global variables
*/
int verbose = FALSE;
int just_read_it = FALSE;
char *progname;
char connection[32] = {
	UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, 
	UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, 
	UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, 
	UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, UNUSED, 
};
char rconnection[32];


#ifndef CONFIGURATION_FILE
#define CONFIGURATION_FILE "vmtm.ncu"
#endif

/*
	With config_array and the connect macro
	the writer of the configuration file builds
	up an array we can use to configure the
	c004.

	Syntax for the configuration file (normally named 
 	vmtm.ncu, see the macro above): One or more lines like

		connect ( ITEM1 , ITEM2 )

	where ITEM1 and ITEM2 are valid link_ids. C-style
	comments and preprocessor statements are also allowed.

	**********************************************
	* connect should only use the link_id macros *
	* from vmtm.h				     *
	**********************************************

	You can define CONFIGURATION_FILE in your makefile. Default
	is "vmtm.ncu".

	For every new configuration you have to call make.
*/
#define connect(c1, c2)		VMTM_C004_CONNECT, c1, c2,
static char config_array[] = {
	VMTM_C004_RESET,
#include CONFIGURATION_FILE
	VMTM_C004_END,
};


/*
	read_configuration reads the contents of the
	c004 into the array named connection. Open links
	are marked with UNUSED.
*/
int read_configuration(link_id, connection)
unsigned char connection[32];
{
	static char read_c004[] = {VMTM_C004_READ, 0};
	int i;
	int lerr;
	for (i=0; i<32; i++) {
		read_c004[1] = i;
		if (WriteLink(link_id, read_c004, 2, WRITE_TIMEOUT) != 2) {
			fprintf(stderr, 
				"%s: can not read configuration from c004",
				progname);
		}
		if ((lerr = ReadLink(link_id, connection+i, 1, READ_TIMEOUT)) 
		    <= 0) {
			/* error/timeout during receive */
			fprintf (stderr,
			     "\n%s: can not read configuration, because %s\n",
			     progname, (lerr == 0) ? "timeout" : lerror(lerr));
			return (!OK);
		}
		if (connection[i] & 0x80)	/* valid value? */
			connection[i] &= ~0x80;	/* yes: yabadabadoo */
		else
			connection[i] = UNUSED;	/* no: link not used */
	}
	return(OK);
}


/*
	print_configuration():
	The contents of the array 'connection' is written to
	stdout. It may look like this:

	-- -- -- -- 21 -- -- -- 22 -- -- -- 23 -- -- --
	-- -- -- -- --  4  8 12 -- -- -- -- -- -- -- --

	The fourth entry (starting with zero) set to 21 means that
	input 4 of the c004 is connected with output 21 and vice versa.
*/

void print_configuration(connection)
char connection[32];
{
	int i;
	for (i=0; i<32; i++) {
		if ((i & 0x0F) == 0)
			printf("\n");
		if (connection[i] != UNUSED)
			printf("%2d ", connection[i]);
		else
			printf("-- ");
	}
	printf("\n");
}


/*
	do_configuration() makes some checks before
	configuring the c004. It also produces a
	connection array similar to the arrays made
	by read_configuration() so that we can proof the
	result.

	return value is !=OK if any error occured!
*/
int do_configuration(link_id)
int link_id;
{
	int i = 0;
	int k;
	int lerr;

	if (verbose) {
		printf("\n%s: check configuration\n", progname);
	}

	/* check for conflict free connections */
	for (k=1, lerr=0; k<sizeof(config_array)-1; k+=3) {
		if (config_array[k] != VMTM_C004_CONNECT) {
		      fprintf(stderr, "\nUnknown command: entry number %d\n\n",
			      k/3+1);
		      exit(1);
		}
		if (connection[config_array[k+1]] != UNUSED) {
			/* redefinition of a connection */
			fprintf(stderr, 
				"\nERROR: double connection to link_id %d\n",
				config_array[k+1]);
			fprintf(stderr, 
				"Entry number %d in configuration file %s\n\n",
				k/3+1, CONFIGURATION_FILE);
			lerr++;
			continue;
		}
		if (connection[config_array[k+2]] != UNUSED) {
			/* redefinition of a connection */
			fprintf(stderr, 
				"\nERROR: double connection to link_id %d\n",
				config_array[k+2]);
			fprintf(stderr, 
				"Entry number %d in configuration file %s\n\n",
				k/3+1, CONFIGURATION_FILE);
			lerr++;
			continue;
		}
		connection[config_array[k+1]] = config_array[k+2];
		connection[config_array[k+2]] = config_array[k+1];
	}

	if (lerr != 0)
		return(!OK);

	if (verbose) {
		printf("%s: configuring\n", progname);
	}
	/* send the configuration to the c004 */
	if ((lerr = WriteLink(link_id, config_array, sizeof(config_array), 
			      CONFIG_TIMEOUT)) 
	    != sizeof(config_array)) {
		/* failed to send complete data */
		if ( lerr >= 0 ) {
			/* timeout during send */
			fprintf (stderr,
				 "\n%s: Timeout! Only %d from %d bytes send\n",
				 progname, lerr, sizeof(config_array));
		} else {
			/* failed to send */
			fprintf (stderr,
				 "\n%s: Failed to send %d bytes, because %s\n",
				 progname,sizeof(config_array), lerror (lerr));
		}
		return (!OK);
	}

	/* read the configuration back */
	if (read_configuration(link_id, rconnection) != OK) {
		return(!OK);
	}

	/* and proof the result */
	for(i = lerr = 0; i<32; i++)
		lerr += (rconnection[i] != connection[i]);
	if (verbose || lerr) {
		print_configuration(rconnection);
		if (lerr)
			printf("%s: %d error(s) occured.\n\n", progname, lerr);
	}
	return((lerr == 0)? OK: !OK);
}



/*	code_load
 *
 *	Resets the transputer and downloads the specified
 *	transputer executable file.
 *	Returns OK on success, the error code from link-interface
 *	otherwise. Error diagnostics are printed directly using stderr.
 *	Return value: OK means SUCCESS, any other value means FAILURE
 */

int code_load (link_id, boot_file)
int link_id;
char *boot_file;
{
	FILE *fp;			/* boot-file */
	int lerr;			/* local error code */
	int cnt;
	int total_send;			/* bytes send so far */
	unsigned char sdata, rdata;	/* send/receive-data for transputer */
	char fbuffer[BUFF_SIZE];	/* boot-file buffer */

	/* reset transputer */
	lerr = ResetLink (link_id);
	if ( lerr < 0 ) {
		/* failed to reset link */
		fprintf (stderr,
			"\n%s: Failed to reset link %d, because %s\n",
			progname, lerror (lerr));
		return (!OK);
	}

	/* download transputer code-file */
	fp = fopen (boot_file, "r");
	if ( fp == NULL ) {
		/* failed to open boot file */
		fprintf (stderr,
			"\n%s: Failed to open boot-file \"%s\", because ",
			progname, boot_file);
		perror ("");
		return(!OK);
	}
	/* boot file opened successfully */
	total_send = 0;
	while ((cnt = fread (fbuffer, 1, sizeof (fbuffer), fp)) > 0) {
		/* try to send fbuffer to transputer */
		lerr = WriteLink (link_id, fbuffer, cnt, BOOT_TIMEOUT);
		if ( lerr != cnt ) {
			/* failed to send complete data */
			if ( lerr >= 0 ) {
				/* timeout during send */
				fprintf (stderr,
				   "\n%s: Only %d from %d bytes send\n",
				   progname, lerr, cnt);
				return(!OK);
			} else {
				/* failed to send */
				fprintf (stderr,
				"\n%s: Failed to send at 0x%X, because %s\n",
				progname, total_send, lerror (lerr));
				return (!OK);
			}
		}
		total_send += cnt;
	} /* while */
	/* check for fread error */
	if ( feof (fp) ) {
		/* end of file */
		fclose (fp);
	} else {
		/* error during read */
		fprintf (stderr,
			"\n%s: Read error from boot-file \"%s\",",
			progname, boot_file);
		perror (" because ");
		return (!OK);
	}
	/* everything is paletti */
	return (OK);
}


int configure(link_name)
char *link_name;
{
	struct ext_link_info *x_link = NULL;
	int no_links;			/* number of links */
	int link_id;			/* used to identify the link */
	int config_type = CONF_NULL;	/* InfoLink() .config_type */
	int lerr;
/*
	open the specified link.
	Rely on LinkId is the index in 'ext_link_info' array.
 */
	if ((no_links = InfoLink (&x_link)) < 0) {
		/* failed to get info about links */
		fprintf (stderr, 
			 "\n%s: Failed to get info about links, because %s\n",
			 progname, lerror (no_links));
		return (EINFOLINK);
	}
	link_id = OpenLink (link_name);
	if ( link_id < 0 ) {
		/* failed to open link */
		fprintf (stderr, 
			 "\n%s: Failed to open link \"%s\", because %s\n",
			 progname, link_name, lerror (link_id));
		return (ELOPEN);
	}

	config_type = x_link[link_id].config_type;
	if (config_type != CONF_VMTM) {
		fprintf(stderr, 
			"\n%s: Link \"%s\" is no VMTM config-link (type %d)\n",
			progname, link_name, config_type);
		return (ENOCONFL);
	}
/*
	now it is sure, that link-name is a config-link for VMTM.
	Reset configuration transputer and download the code_file.
	Read or do the configuration and close the configuration link.
*/
	if (code_load (link_id, VMTM_BOOT_FILE) == OK) {
		if (just_read_it) {
			if (verbose) {
				printf("\n%s: read configuration of %s\n", 
				       progname, link_name);
			}
			read_configuration(link_id, rconnection);
			print_configuration(rconnection);
			lerr = OK;
		} else {
			lerr = do_configuration(link_id);
		}
	}
	CloseLink (link_id);
	return (lerr);
}


void usage()
{
	printf ("\nusage: %s [-v | -r] link-name\n%s\n%s\n%s\n",
		progname,
		"configure Parsytec\'s VMTM transputer board",
		"  -v   verbose",
		"  -r   do not configure, display existing topology");
	fflush (stdout);
}


int main(argc, argv)
int argc;
char *argv[];
{
	char *link_name = NULL;
	int lerr;

	progname = argv[0];
	while (--argc > 0) {
		if (**++argv == '-') {
			switch (argv[0][1]) {
				case 'v': verbose = TRUE;
				          break;
				case 'r': just_read_it = TRUE;
				          verbose = TRUE;
				          break;
				default:  usage();
				          exit(EPARAM);
			}
		} else {
			if (link_name == NULL) {
				link_name = *argv;
			} else {
				usage();
				exit(EPARAM);
			}
		}
	}
	if (link_name == NULL) {
		usage();
		exit(EPARAM);
	}

	lerr = configure (link_name);
	return (lerr != OK);
}



